@@ -4,7 +4,7 @@
 
 ;; Author: Jim Porter
 ;; URL: https://github.com/jimporter/urgrep
-;; Version: 0.1.1
+;; Version: 0.2.0
 ;; Keywords: grep, search
 ;; Package-Requires: ((emacs "27.1") (compat "29.1.0.1") (project "0.3.0"))
 
@@ -37,8 +37,9 @@
 (require 'generator)
 (require 'grep)
 (require 'project)
+(require 'shell)                        ; For `shell--parse-pcomplete-arguments'
 (require 'text-property-search)
-(require 'xref)                         ; For `xref--regexp-to-extended'.
+(require 'xref)                         ; For `xref--regexp-to-extended'
 
 (defgroup urgrep nil
   "Run a grep-like command and display the results."
@@ -80,6 +81,11 @@ Valid values are nil (case-sensitive), t (case-insensitive), `smart'
                 (const :tag "Case insensitive" t))
   :group 'urgrep)
 
+(defcustom urgrep-search-hidden-files nil
+  "If non-nil, default to searching in hidden files."
+  :type 'boolean
+  :group 'urgrep)
+
 (defcustom urgrep-context-lines 0
   "Number of lines of context to show.
 If this is an integer, show that many lines of context on either
@@ -114,6 +120,16 @@ The currently-used tool can be inspected from the hook via
 
 ;; Urgrep utility functions
 
+;; `format-prompt' was added in Emacs 28.1.
+(defalias 'urgrep--split-string-shell-command
+  (if (fboundp 'split-string-shell-command)
+      #'split-string-shell-command
+    (lambda (string)
+      (with-temp-buffer
+        (insert string)
+        (let ((comint-file-name-quote-list shell-file-name-quote-list))
+          (car (shell--parse-pcomplete-arguments)))))))
+
 (defun urgrep--convert-regexp (expr from-syntax to-syntax)
   "Convert the regexp EXPR from FROM-SYNTAX to TO-SYNTAX."
   (cond ((and (not (eq from-syntax to-syntax))
@@ -196,13 +212,6 @@ one for each `:abbreviate' key found."
     (when tree (push (urgrep--maybe-shell-quote-argument tree) elems))
     (string-join (nreverse elems) " ")))
 
-(defmacro urgrep--with-killed-local-variable (variable &rest body)
-  "Execute the forms in BODY with VARIABLE temporarily non-local."
-  (declare (indent 1))
-  `(if (local-variable-p ,variable)
-       (with-temp-buffer ,@body)
-     ,@body))
-
 (rx-define urgrep-regular-number (seq (any "1-9") (* digit)))
 
 
@@ -215,16 +224,20 @@ one for each `:abbreviate' key found."
     ((or `(,c . ,c) (and c (pred numberp))) (list (format "-C%d" c)))
     (`(,b . ,a) (list (format "-B%d" b) (format "-A%d" a)))))
 
-(cl-defun urgrep--rgrep-command (query &key tool regexp case-fold files context
-                                       color &allow-other-keys)
+(cl-defun urgrep--rgrep-command (query &key tool regexp case-fold hidden files
+                                       directory context color
+                                       &allow-other-keys)
   "Get the command to run for QUERY when using rgrep.
-Optional keys TOOL, REGEXP, CASE-FOLD, FILES, CONTEXT, and COLOR are
-as in `urgrep-command'."
+Optional keys TOOL, REGEXP, CASE-FOLD, HIDDEN, FILES, DIRECTORY,
+CONTEXT, and COLOR are as in `urgrep-command'."
   (grep-compute-defaults)
   ;; Locally add options to `grep-find-template' that grep.el isn't aware of.
   (let ((grep-find-template grep-find-template)
         (grep-highlight-matches (if color 'always nil))
-        (files (if files (mapconcat #'identity files " ") "*")))
+        (files (if files (mapconcat #'identity files " ") "*"))
+        (directory (when directory
+                     (mapconcat #'urgrep--maybe-shell-quote-argument
+                                directory " "))))
     (pcase-dolist (`(,k . ,v) `((regexp    . ,regexp)
                                 (case-fold . ,case-fold)
                                 (context   . ,context)))
@@ -234,16 +247,59 @@ as in `urgrep-command'."
                  ((string-match "<C>" grep-find-template)))
         (setq grep-find-template
               (replace-match (concat "<C> " args) t t grep-find-template))))
-    (let* ((case-fold-search nil)
-           (command (rgrep-default-command query files nil)))
-      (save-match-data
-        ;; Hide excessive part of rgrep command.
-        (when (string-match
-               (rx bol "find " (group (*? nonl)) " " (or "-exec" "-print"))
-               command)
-          (put-text-property (match-beginning 1) (match-end 1)
-                             'abbreviated-command t command)))
-      command)))
+    (let ((case-fold-search nil)
+          (grep-find-ignored-directories grep-find-ignored-directories)
+          (grep-find-ignored-files grep-find-ignored-files))
+      (unless hidden
+        (setq grep-find-ignored-directories
+              (cons ".*" (seq-filter (lambda (s) (not (string-prefix-p "." s)))
+                                     grep-find-ignored-directories))
+              grep-find-ignored-files
+              (cons ".*" (seq-filter (lambda (s) (not (string-prefix-p "." s)))
+                                     grep-find-ignored-files))))
+      (let ((command (rgrep-default-command query files directory)))
+        (save-match-data
+          ;; Hide excessive part of rgrep command.
+          (when (string-match
+                 (rx bol "find " (group (*? nonl)) " " (or "-exec" "-print"))
+                 command)
+            (put-text-property (match-beginning 1) (match-end 1)
+                               'abbreviated-command t command)))
+        command))))
+
+(cl-defun urgrep--git-grep-command (query &key tool regexp case-fold hidden
+                                          files directory group context color)
+  "Get the command to run for QUERY when using git grep.
+Optional keys TOOL, REGEXP, CASE-FOLD, HIDDEN, FILES, DIRECTORY,
+CONTEXT, and COLOR are as in `urgrep-command'."
+  ;; XXX: This is very similar to the implementation in `urgrep-command', except
+  ;; that we do some extra work to generate pathspecs. Can we factor out some of
+  ;; this?
+  (let* ((arguments (urgrep--get-prop 'arguments tool))
+         (abbrev (urgrep--get-prop 'abbreviations tool))
+         (pathspecs
+          (if (and files directory)
+              (mapcan
+               (lambda (file)
+                 (mapcar (lambda (dir)
+                           (concat ":(glob)" (file-name-concat dir "**" file)))
+                         directory))
+               files)
+            (or files directory)))
+         (props `((executable . ,(urgrep--get-prop 'executable-name tool))
+                  (query . ,query)
+                  ,@(mapcar (pcase-lambda (`(,k . ,v))
+                              (cons k (urgrep--get-prop-pcase
+                                       k tool v "-arguments")))
+                            `((regexp         . ,regexp)
+                              (case-fold      . ,case-fold)
+                              (hidden-file    . ,hidden)
+                              (pathspec       . ,pathspecs)
+                              (group          . ,group)
+                              (context        . ,context)
+                              (color          . ,color))))))
+    (urgrep--flatten-arguments (cl-sublis props arguments)
+                               abbrev)))
 
 (defun urgrep--rgrep-process-setup ()
   "Set up environment variables for rgrep.
@@ -260,16 +316,19 @@ See also `grep-process-setup'."
   `((ugrep
      (executable-name . "ugrep")
      (regexp-syntax bre ere pcre)
-     (arguments executable (:abbreviate color "-n" "--ignore-files")
-                file-wildcards group context case-fold regexp "-e" query)
+     (arguments executable (:abbreviate color "-rn" "--ignore-files")
+                hidden-file file-wildcards group context case-fold regexp "-e"
+                query directory)
      (regexp-arguments ('bre  '("-G"))
                        ('ere  '("-E"))
                        ('pcre '("-P"))
                        (_     '("-F")))
      (case-fold-arguments ((pred identity) '("-i")))
+     (hidden-file-arguments ((pred identity) '("--hidden")))
      (file-wildcards-arguments
       ((and x (pred identity))
        (mapcar (lambda (i) (concat "--include=" i)) x)))
+     (directory-arguments (x x))
      (group-arguments ((pred identity) '("--heading" "--break")))
      (context-arguments . ,urgrep--context-arguments)
      (color-arguments
@@ -279,13 +338,15 @@ See also `grep-process-setup'."
     (ripgrep
      (executable-name . "rg")
      (regexp-syntax pcre)
-     (arguments executable (:abbreviate color) file-wildcards group context
-                case-fold regexp "--" query)
+     (arguments executable (:abbreviate color) hidden-file file-wildcards group
+                context case-fold regexp "--" query directory)
      (regexp-arguments ('nil '("-F")))
      (case-fold-arguments ((pred identity) '("-i")))
+     (hidden-file-arguments ((pred identity) '("--hidden")))
      (file-wildcards-arguments
       ((and x (pred identity))
        (flatten-list (mapcar (lambda (i) (cons "-g" i)) x))))
+     (directory-arguments (x x))
      (group-arguments ('nil '("--no-heading"))
                       (_    '("--heading")))
      (context-arguments . ,urgrep--context-arguments)
@@ -298,14 +359,16 @@ See also `grep-process-setup'."
     (ag
      (executable-name . "ag")
      (regexp-syntax pcre)
-     (arguments executable (:abbreviate color) file-wildcards group context
-                case-fold regexp "--" query)
+     (arguments executable (:abbreviate color) hidden-file file-wildcards group
+                context case-fold regexp "--" query directory)
      (regexp-arguments ('nil '("-Q")))
      (case-fold-arguments ('nil '("-s"))
                           (_    '("-i")))
+     (hidden-file-arguments ((pred identity) '("--hidden")))
      (file-wildcards-arguments
       ((and x (pred identity))
        (list "-G" (urgrep--wildcards-to-regexp x 'pcre))))
+     (directory-arguments (x x))
      (group-arguments ('nil '("--nogroup"))
                       (_    '("--group")))
      (context-arguments . ,urgrep--context-arguments)
@@ -315,13 +378,16 @@ See also `grep-process-setup'."
     (ack
      (executable-name . "ack")
      (regexp-syntax pcre)
-     (arguments executable (:abbreviate color) file-wildcards group context
-                case-fold regexp "--" query)
+     (arguments executable (:abbreviate color) hidden-file file-wildcards group
+                context case-fold regexp "--" query directory)
      (regexp-arguments ('nil '("-Q")))
      (case-fold-arguments ((pred identity) '("-i")))
+     (hidden-file-arguments ('nil '("--ignore-dir=match:/^\\./"
+                                    "--ignore-file=match:/^\\./")))
      (file-wildcards-arguments
       ((and x (pred identity))
        (list "-G" (urgrep--wildcards-to-regexp x 'pcre))))
+     (directory-arguments (x x))
      (group-arguments ('nil '("--nogroup"))
                       (_    '("--group")))
      (context-arguments . ,urgrep--context-arguments)
@@ -336,16 +402,19 @@ See also `grep-process-setup'."
      ;; with people who want to customize the arguments.
      (vc-backend . "Git")
      (regexp-syntax bre ere pcre)
+     (command-function . ,#'urgrep--git-grep-command)
      (arguments executable (:abbreviate "--no-pager" color "--no-index"
                                         "--exclude-standard" "-n")
-                group context case-fold regexp "-e" query "--" file-wildcards)
+                group context case-fold regexp "-e" query "--" hidden-file
+                pathspec)
      (abbreviations "grep")
      (regexp-arguments ('bre  '("-G"))
                        ('ere  '("-E"))
                        ('pcre '("-P"))
                        (_     '("-F")))
      (case-fold-arguments ((pred identity) '("-i")))
-     (file-wildcards-arguments (x x))
+     (hidden-file-arguments ('nil '(":!.*")))
+     (pathspec-arguments (x x))
      (group-arguments ((pred identity) '("--heading" "--break")))
      (context-arguments . ,urgrep--context-arguments)
      ;; git is a bit odd in that color specification happens *before* the
@@ -362,7 +431,9 @@ See also `grep-process-setup'."
                               "-c" "color.grep.separator="
                               "grep" "--color"))))
     (grep
-     (executable-name . "grep")
+     ;; Note: We only use these for detecting the usability of find/grep. To
+     ;; modify the programs that actually run, change `grep-find-template'.
+     (executable-name "find" "grep")
      (regexp-syntax bre ere pcre)
      (command-function . ,#'urgrep--rgrep-command)
      (process-setup . ,#'urgrep--rgrep-process-setup)
@@ -380,9 +451,10 @@ See also `grep-process-setup'."
 (defcustom urgrep-preferred-tools nil
   "List of urgrep tools to search for.
 This can be nil to use the default list of tools in `urgrep-tools'
-or a list of tools to try in descending order of preference.  Each tool
-can be either a symbol naming the tool or a cons cell of the tool name
-and the path of the executable."
+or a list of tools to try in descending order of preference. Each
+tool can be either a symbol naming the tool or a cons cell of the
+tool name and the file name of the executable (or a list thereof
+if there are multiple exeuctables)."
   :type `(choice
           (const :tag "Default" nil)
           (repeat :tag "List of tools"
@@ -461,7 +533,8 @@ This caches the default tool per-host in `urgrep--host-defaults'."
                                      (project-root proj))))
             ;; If we find the executable (and it's for the right VC
             ;; backend, if relevant), cache it and then return it.
-            (when (and (executable-find tool-executable t)
+            (when (and (seq-every-p (lambda (i) (executable-find i t))
+                                    (ensure-list tool-executable))
                        (or (not tool-vc-backend)
                            (string= vc-backend-name tool-vc-backend)))
               ;; So long as we didn't examine a VC-specific tool, we can
@@ -482,13 +555,29 @@ This caches the default tool per-host in `urgrep--host-defaults'."
 
 (defun urgrep-get-tool (&optional tool)
   "Get the urgrep tool for TOOL.
-If TOOL is nil, get the default tool.  If TOOL is a string, look it up
+If TOOL is nil, get the default tool.  If TOOL is a symbol, look it up
 in `urgrep-tools'.  Otherwise, return TOOL as-is."
   (pcase tool
     ('nil (urgrep--get-default-tool))
     ((and (pred symbolp) tool) (assq tool urgrep-tools))
     (tool tool)))
 
+(defun urgrep--guess-tool (command)
+  "Guess the urgrep tool from the specified COMMAND."
+  (catch 'found
+    (when-let ((args (urgrep--split-string-shell-command
+                      ;; First, split by semicolon, since
+                      ;; `split-string-shell-command' only returns the *last*
+                      ;; command.
+                      (car (split-string command ";"))))
+               (command-name (file-name-nondirectory (car args))))
+      (dolist (tool urgrep-tools)
+        (when (string= command-name
+                       (car (ensure-list (urgrep--get-prop
+                                          'executable-name tool))))
+          (throw 'found tool))))
+    (error "Unable to guess urgrep tool from command")))
+
 (defun urgrep--get-best-syntax (syntax tool)
   "Return the regexp syntax closest to SYNTAX that TOOL supports."
   (let ((tool-syntaxes (urgrep--get-prop 'regexp-syntax tool)))
@@ -499,9 +588,8 @@ in `urgrep-tools'.  Otherwise, return TOOL as-is."
           (t (car tool-syntaxes)))))
 
 ;;;###autoload
-(cl-defun urgrep-command (query &key tool regexp (case-fold 'inherit) files
-                                (group t) (context 0) (color t)
-                                (directory default-directory))
+(cl-defun urgrep-command (query &key tool regexp (case-fold 'inherit) hidden
+                                files directory (group t) (context 0) (color t))
   "Return a command to use to search for QUERY.
 Several keyword arguments can be supplied to adjust the resulting
 command:
@@ -517,6 +605,8 @@ strings), `bre' (basic regexp), `ere' (extend regexp), `pcre'
 CASE-FOLD: determine whether QUERY is case-sensitive or not; possible
 values are as `urgrep-case-fold', defaulting to `inherit'.
 
+HIDDEN: non-nil to search in hidden files; defaults to nil.
+
 FILES: a wildcard (or list of wildcards) to limit the files searched.
 
 GROUP: show results grouped by filename (t, the default), or if nil,
@@ -526,44 +616,44 @@ CONTEXT: the number of lines of context to show around results; either
 an integer (to show the same number of lines before and after) or a
 cons (to show CAR and CDR lines before and after, respectively).
 
-COLOR: non-nil (the default) if the output should use color.
-
-DIRECTORY: the directory to search in, or nil to use the
-`default-directory'."
-  (let ((default-directory (or directory default-directory)))
-    (with-connection-local-variables
-     (let* ((regexp-syntax (if (eq regexp t) urgrep-regexp-syntax regexp))
-            (files (if (listp files) files (list files)))
-            (tool (or (urgrep-get-tool tool)
-                      (error "unknown tool %s" tool)))
-            (tool-re-syntax (urgrep--get-best-syntax regexp-syntax tool))
-            (query (urgrep--convert-regexp query regexp-syntax tool-re-syntax))
-            (cmd-fun (urgrep--get-prop 'command-function tool)))
-       ;; Determine whether to search case-sensitively or not.
-       (when (eq case-fold 'inherit)
-         (setq case-fold (if case-fold-search 'smart nil)))
-       (when (eq case-fold 'smart)
-         (setq case-fold (isearch-no-upper-case-p query regexp-syntax)))
-       ;; Build the command arguments.
-       (if cmd-fun
-           (funcall cmd-fun query :tool tool :regexp regexp-syntax
-                    :case-fold case-fold :files files :group group
-                    :context context :color color)
-         (let ((arguments (urgrep--get-prop 'arguments tool))
-               (abbrev (urgrep--get-prop 'abbreviations tool))
-               (props `((executable . ,(urgrep--get-prop 'executable-name tool))
-                        (query . ,query)
-                        ,@(mapcar (pcase-lambda (`(,k . ,v))
-                                    (cons k (urgrep--get-prop-pcase
-                                             k tool v "-arguments")))
-                                  `((regexp         . ,tool-re-syntax)
-                                    (case-fold      . ,case-fold)
-                                    (file-wildcards . ,files)
-                                    (group          . ,group)
-                                    (context        . ,context)
-                                    (color          . ,color))))))
-           (urgrep--flatten-arguments (cl-sublis props arguments)
-                                      abbrev)))))))
+COLOR: non-nil (the default) if the output should use color."
+  (with-connection-local-variables
+   (let* ((regexp-syntax (if (eq regexp t) urgrep-regexp-syntax regexp))
+          (files (if (listp files) files (list files)))
+          (directory (if (listp directory) directory (list directory)))
+          (tool (or (urgrep-get-tool tool)
+                    (error "unknown tool %s" tool)))
+          (tool-re-syntax (urgrep--get-best-syntax regexp-syntax tool))
+          (query (urgrep--convert-regexp query regexp-syntax tool-re-syntax))
+          (cmd-fun (urgrep--get-prop 'command-function tool)))
+     ;; Determine whether to search case-sensitively or not.
+     (when (eq case-fold 'inherit)
+       (setq case-fold (if case-fold-search 'smart nil)))
+     (when (eq case-fold 'smart)
+       (setq case-fold (isearch-no-upper-case-p query regexp-syntax)))
+     ;; Build the command arguments.
+     (if cmd-fun
+         (funcall cmd-fun query :tool tool :regexp tool-re-syntax
+                  :case-fold case-fold :hidden hidden :files files
+                  :directory directory :group group :context context
+                  :color color)
+       (let ((arguments (urgrep--get-prop 'arguments tool))
+             (abbrev (urgrep--get-prop 'abbreviations tool))
+             (props `((executable . ,(urgrep--get-prop 'executable-name tool))
+                      (query . ,query)
+                      ,@(mapcar (pcase-lambda (`(,k . ,v))
+                                  (cons k (urgrep--get-prop-pcase
+                                           k tool v "-arguments")))
+                                `((regexp         . ,tool-re-syntax)
+                                  (case-fold      . ,case-fold)
+                                  (hidden-file    . ,hidden)
+                                  (file-wildcards . ,files)
+                                  (directory      . ,directory)
+                                  (group          . ,group)
+                                  (context        . ,context)
+                                  (color          . ,color))))))
+         (urgrep--flatten-arguments (cl-sublis props arguments)
+                                    abbrev))))))
 
 
 ;; urgrep-mode
@@ -595,30 +685,27 @@ If EDIT-COMMAND is non-nil, the search can be edited."
                     query)))
     (urgrep--start command query urgrep-current-tool)))
 
-(defvar urgrep-mode-map
-  (let ((map (make-sparse-keymap)))
-    ;; Don't inherit from `compilation-minor-mode-map',
-    ;; because that introduces a menu bar item we don't want.
-    (set-keymap-parent map special-mode-map)
-    (define-key map [mouse-2] #'compile-goto-error)
-    (define-key map [follow-link] 'mouse-face)
-    (define-key map "\C-c\C-c" #'compile-goto-error)
-    (define-key map "\C-m" #'compile-goto-error)
-    (define-key map "\C-o" #'compilation-display-error)
-    (define-key map "\C-c\C-k" #'kill-compilation)
-    (define-key map "\M-n" #'compilation-next-error)
-    (define-key map "\M-p" #'compilation-previous-error)
-    (define-key map "\M-{" #'compilation-previous-file)
-    (define-key map "\M-}" #'compilation-next-file)
-    (define-key map "n" #'next-error-no-select)
-    (define-key map "p" #'previous-error-no-select)
-    (define-key map "{" #'compilation-previous-file)
-    (define-key map "}" #'compilation-next-file)
-    (define-key map "\t" #'compilation-next-error)
-    (define-key map [backtab] #'compilation-previous-error)
-    (define-key map "g" #'urgrep-search-again)
-    map)
-  "Keymap for urgrep buffers.")
+(defvar-keymap urgrep-mode-map
+  ;; Don't inherit from `compilation-minor-mode-map',
+  ;; because that introduces a menu bar item we don't want.
+  :parent special-mode-map
+  "<mouse-2>"     #'compile-goto-error
+  "<follow-link>" 'mouse-face
+  "C-c C-c"       #'compile-goto-error
+  "C-m"           #'compile-goto-error
+  "C-o"           #'compilation-display-error
+  "C-c C-k"       #'kill-compilation
+  "M-n"           #'compilation-next-error
+  "M-p"           #'compilation-previous-error
+  "M-{"           #'compilation-previous-file
+  "M-}"           #'compilation-next-file
+  "n"             #'next-error-no-select
+  "p"             #'previous-error-no-select
+  "{"             #'compilation-previous-file
+  "}"             #'compilation-next-file
+  "TAB"           #'compilation-next-error
+  "<backtab>"     #'compilation-previous-error
+  "g"             #'urgrep-search-again)
 
 (easy-menu-define urgrep-menu-map urgrep-mode-map
   "Menu for urgrep buffers."
@@ -661,13 +748,10 @@ If EDIT-COMMAND is non-nil, the search can be edited."
        :help "Restart search")
       map)))
 
-(defvar urgrep-mode-abbreviation-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [down-mouse-2] 'mouse-set-point)
-    (define-key map [mouse-2] 'grep-find-toggle-abbreviation)
-    (define-key map "\C-m" 'grep-find-toggle-abbreviation)
-    map)
-  "Keymap for urgrep abbreviation buttons.")
+(defvar-keymap urgrep-mode-abbreviation-map
+  "<down-mouse-2>" #'mouse-set-point
+  "<mouse-2>"      #'grep-find-toggle-abbreviation
+  "C-m"            #'grep-find-toggle-abbreviation)
 
 (defconst urgrep-mode-line-matches
   `(" [" (:propertize (:eval (int-to-string urgrep-num-matches-found))
@@ -864,13 +948,13 @@ This function is called from `compilation-filter-hook'."
 (define-compilation-mode urgrep-mode "Urgrep"
   "A compilation mode for various grep-like tools."
   (setq-local tool-bar-map urgrep-mode-tool-bar-map
-              compilation-process-setup-function 'urgrep-process-setup
+              compilation-process-setup-function #'urgrep-process-setup
               compilation-error-face 'urgrep-hit
               compilation-error-regexp-alist urgrep-regexp-alist
               compilation-mode-line-errors urgrep-mode-line-matches
               compilation-disable-input t
               compilation-error-screen-columns nil)
-  (add-hook 'compilation-filter-hook 'urgrep-filter nil t))
+  (add-hook 'compilation-filter-hook #'urgrep-filter nil t))
 
 (defun urgrep--hide-abbreviations (command)
   "If `urgrep-abbreviate-command' is non-nil, hide abbreviations in COMMAND."
@@ -904,9 +988,10 @@ rerunning the search."
                       (file-name-as-directory (expand-file-name directory))
                     default-directory))
   (with-current-buffer
-      ;; Dynamically bind `urgrep-current-tool' so that `urgrep-process-filter'
-      ;; can consult it.
-      (urgrep--with-killed-local-variable 'urgrep-current-tool
+      ;; Run this in a temporary buffer to ensure that we let-bind the default
+      ;; binding for `urgrep-current-tool'.  `urgrep-process-setup' needs to
+      ;; consult this let-binding from another buffer.
+      (with-temp-buffer
         ;; Let-bind `default-directory' here so that the external command knows
         ;; where to search...
         (let ((urgrep-current-tool tool)
@@ -970,6 +1055,23 @@ indicates the default query, if any."
   ;; was at the beginning of the editable text.
   (if (< (point) (minibuffer-prompt-end)) (goto-char (minibuffer-prompt-end))))
 
+(defun urgrep-set-tool (tool)
+  "Set the TOOL to use for the current search.
+TOOL should be a symbol corresponding to one of the keys in
+`urgrep-tools'."
+  (interactive
+   (let* ((enable-recursive-minibuffers t)
+          (tool-string (completing-read
+                        (format-prompt "Tool" (car urgrep-current-tool))
+                        (mapcar (lambda (i) (symbol-name (car i)))
+                                urgrep-tools)
+                        nil 'require-match nil nil
+                        (symbol-name (car urgrep-current-tool)))))
+     (list (unless (string= tool-string "") (intern tool-string)))))
+  (when tool
+    (setq-default urgrep-current-tool (urgrep-get-tool tool))
+    (message "using %s" tool)))
+
 (defun urgrep-toggle-regexp ()
   "Toggle whether or not to use regexps for the search query.
 Within the `urgrep' search prompt, this sets the value only for the
@@ -1031,6 +1133,17 @@ future searches."
     (setq urgrep-context-lines (cons before-lines after-lines)))
   (when (window-minibuffer-p) (urgrep--update-search-prompt)))
 
+(defun urgrep-toggle-search-hidden-files ()
+  "Toggle whether or not to search in hidden files.
+Within the `urgrep' search prompt, this sets the value only for the
+current search.  Outside the prompt, this sets the value for all
+future searches."
+  (interactive)
+  (message (if (setq urgrep-search-hidden-files
+                     (not urgrep-search-hidden-files))
+               "search hidden"
+             "exclude hidden")))
+
 (defun urgrep-set-file-wildcards (files)
   "Set the FILES (a wildcard or list thereof) to search.
 Within the `urgrep' search prompt, this sets the value only for the
@@ -1042,42 +1155,54 @@ future searches."
   (setq urgrep-file-wildcards files)
   (when (window-minibuffer-p) (urgrep--update-search-prompt)))
 
-(defvar urgrep-minibuffer-map
-  (let ((map (make-sparse-keymap)))
-    (set-keymap-parent map minibuffer-local-map)
-    (define-key map "\M-sr" #'urgrep-toggle-regexp)
-    (define-key map "\M-sc" #'urgrep-toggle-case-fold)
-    (define-key map "\M-sf" #'urgrep-set-file-wildcards)
-    (define-key map "\M-sC" #'urgrep-set-context)
-    (define-key map "\M-sB" #'urgrep-set-before-context)
-    (define-key map "\M-sA" #'urgrep-set-after-context)
-    map))
+(defvar-keymap urgrep-minibuffer-map
+  :parent minibuffer-local-map
+  "M-s r" #'urgrep-toggle-regexp
+  "M-s c" #'urgrep-toggle-case-fold
+  "M-s f" #'urgrep-set-file-wildcards
+  "M-s h" #'urgrep-toggle-search-hidden-files
+  "M-s C" #'urgrep-set-context
+  "M-s B" #'urgrep-set-before-context
+  "M-s A" #'urgrep-set-after-context
+  "M-s t" #'urgrep-set-tool)
 
 (cl-defun urgrep--read-query (initial &key tool (regexp urgrep-search-regexp)
                                       (case-fold urgrep-case-fold)
+                                      (hidden urgrep-search-hidden-files)
                                       (files urgrep-file-wildcards)
                                       (group urgrep-group-matches)
                                       (context urgrep-context-lines)
-                                      (directory default-directory))
+                                      (default (and (not initial)
+                                                    (urgrep--search-default))))
   "Prompt the user for a search query starting with an INITIAL value.
-Return a list that can be passed to `urgrep-command' to turn into a shell
-command.  TOOL, REGEXP, CASE-FOLD, FILES, GROUP, CONTEXT, and DIRECTORY
- are as in `urgrep-command'."
-  (let* ((default-directory directory)
-         (urgrep-search-regexp regexp)
-         (urgrep-case-fold case-fold)
-         (urgrep-file-wildcards files)
-         (urgrep-context-lines context)
-         (default (and (not initial) (urgrep--search-default)))
-         (prompt (urgrep--search-prompt default))
-         (query (minibuffer-with-setup-hook
-                    (lambda () (setq-local urgrep--search-default default))
-                  (read-from-minibuffer prompt initial urgrep-minibuffer-map nil
-                                        'urgrep-search-history default)))
-         (query (if (equal query "") default query)))
-    (list query :tool (urgrep-get-tool tool) :regexp urgrep-search-regexp
-          :case-fold urgrep-case-fold :files urgrep-file-wildcards :group group
-          :context urgrep-context-lines :directory directory)))
+Return a list that can be passed to `urgrep-command' to turn into
+a shell command. TOOL, REGEXP, CASE-FOLD, FILES, GROUP, and
+CONTEXT are as in `urgrep-command'.
+
+In addition, you can pass DEFAULT to specify the default search
+query. If nil, guess the default based on the current region or
+point."
+  ;; Run this in a temporary buffer to make sure that none of the dynamic
+  ;; variables below that we let-bind have buffer-local bindings. If they did,
+  ;; we wouldn't be able to retrieve the values for them that we set from inside
+  ;; the minibuffer.
+  (with-temp-buffer
+    (let* ((urgrep-current-tool (urgrep-get-tool tool))
+           (urgrep-search-regexp regexp)
+           (urgrep-case-fold case-fold)
+           (urgrep-search-hidden-files hidden)
+           (urgrep-file-wildcards files)
+           (urgrep-context-lines context)
+           (prompt (urgrep--search-prompt default))
+           (query (minibuffer-with-setup-hook
+                      (lambda () (setq-local urgrep--search-default default))
+                    (read-from-minibuffer prompt initial urgrep-minibuffer-map
+                                          nil 'urgrep-search-history default)))
+           (query (if (equal query "") default query)))
+      (list query :tool urgrep-current-tool :regexp urgrep-search-regexp
+            :case-fold urgrep-case-fold :hidden urgrep-search-hidden-files
+            :files urgrep-file-wildcards :group group
+            :context urgrep-context-lines))))
 
 (defun urgrep--read-command (command)
   "Read a shell command to use for searching, with initial value COMMAND."
@@ -1121,29 +1246,42 @@ Type \\[urgrep-set-context] to set the number of context lines.
   lines.  Without a prefix, prompt for the number.
 Type \\[urgrep-set-before-context] to set the number of before context lines.
 Type \\[urgrep-set-after-context] to set the number of after context lines.
+Type \\[urgrep-toggle-search-hidden-files] to toggle searching in hidden files.
 Type \\[urgrep-set-file-wildcards] to set a wildcard to filter the files \
 searched."
   (interactive
-   (let ((directory (urgrep--read-directory current-prefix-arg)))
-     (urgrep--read-query nil :directory directory)))
-  (let* ((full-query (cons query rest))
-         (command (apply #'urgrep-command full-query))
-         (tool (urgrep-get-tool (cadr (cl-member :tool full-query))))
-         (directory (cadr (cl-member :directory full-query))))
+   (let ((default-directory (urgrep--read-directory current-prefix-arg)))
+     `(,@(urgrep--read-query nil) :default-directory ,default-directory)))
+  (let* ((tool (urgrep-get-tool (plist-get rest :tool)))
+         (directory (prog1 (plist-get rest :default-directory)
+                      (cl-remf rest :default-directory)))
+         (full-query (cons query rest))
+         (command (apply #'urgrep-command full-query)))
     (urgrep--start command full-query tool directory)))
 
 ;;;###autoload
-(defun urgrep-run-command (command directory tool)
-  "Recursively search in DIRECTORY using the given COMMAND.
+(cl-defun urgrep-run-command (command &key ((:default-directory directory))
+                                      tool)
+  "Recursively search using the given COMMAND.
+
+TOOL is the search tool corresponding to COMMAND (see
+`urgrep-get-tool'). If nil, guess the tool based on the value of
+COMMAND.
 
 When called interactively, this behaves like `urgrep', but allows you
-to edit the command before running it."
+to edit the command before running it.
+
+\(fn COMMAND &key TOOL)"
   (interactive
-   (let* ((directory (urgrep--read-directory current-prefix-arg))
-          (query (urgrep--read-query nil :directory directory)))
-     (list (urgrep--read-command (apply #'urgrep-command query))
-           directory (cadr (cl-member :tool query)))))
-  (urgrep--start command command (urgrep-get-tool tool) directory))
+   (let* ((default-directory (urgrep--read-directory current-prefix-arg))
+          (query (urgrep--read-query nil))
+          (command (urgrep--read-command (apply #'urgrep-command query)))
+          (tool (condition-case nil (urgrep--guess-tool command)
+                  (error (plist-get (cdr query) :tool)))))
+     (list command :default-directory default-directory :tool tool)))
+  (let ((tool (if tool (urgrep-get-tool tool)
+                (urgrep--guess-tool command))))
+    (urgrep--start command command tool directory)))
 
 (cl-eval-when (compile)
   (require 'esh-cmd)
@@ -1176,6 +1314,11 @@ This is meant to be used as a command in Eshell."
          "ignore case when searching for PATTERN")
      (?S "smart-case" nil case-fold
          "ignore case when searching for PATTERN if PATTERN is all lower case")
+     ;; Hidden file options
+     (nil "hidden" (t) hidden
+          "search hidden files")
+     (nil "no-hidden" (nil) hidden
+          "don't search hidden files")
      ;; Grouping options
      (nil "group" (t) group
           "group results by file")
@@ -1190,36 +1333,34 @@ This is meant to be used as a command in Eshell."
          "number of lines of trailing context to print")
      ;; General options
      (?h "help" nil nil "show this help message")
-     :usage "[OPTION]... PATTERN [PATH]
+     :usage "[OPTION]... PATTERN [PATH]...
 Recursively search for PATTERN within PATH.")
    (unless args (error "Expected a search pattern"))
-   ;; TODO: Add support for multiple paths.
-   (when (> (length args) 2) (error "Only one path supported currently"))
-   (let ((query (car args))
-         (directory (or (cadr args) default-directory))
-         (context
-          (cond (context-around (string-to-number context-around))
-                ((or context-before context-after)
-                 (cons (if context-before (string-to-number context-before) 0)
-                       (if context-after (string-to-number context-after) 0)))))
-         options)
+   (let* ((query (car args))
+          (directory (cdr args))
+          (context
+           (cond
+            (context-around (string-to-number context-around))
+            ((or context-before context-after)
+             (cons (if context-before (string-to-number context-before) 0)
+                   (if context-after (string-to-number context-after) 0)))))
+          options)
      ;; Fill the options to pass to `urgrep'.
-     (when context (setq options (nconc `(:context ,context) options)))
-     (when group (setq options (nconc `(:group ,(car group)) options)))
-     (when case-fold (setq options (nconc `(:case-fold ,(car case-fold))
-                                          options)))
-     (when regexp (setq options (nconc `(:regexp ,(car regexp)) options)))
+     (when directory (setq options `(:directory ,directory       . ,options)))
+     (when context   (setq options `(:context   ,context         . ,options)))
+     (when hidden    (setq options `(:hidden    ,(car hidden)    . ,options)))
+     (when group     (setq options `(:group     ,(car group)     . ,options)))
+     (when case-fold (setq options `(:case-fold ,(car case-fold) . ,options)))
+     (when regexp    (setq options `(:regexp    ,(car regexp)    . ,options)))
      ;; Run `urgrep'.
      (if (and (not (bound-and-true-p eshell-plain-grep-behavior))
               (eshell-interactive-output-p)
               (not eshell-in-pipeline-p)
               (not eshell-in-subcommand-p))
-         (apply #'urgrep query :directory directory options)
+         (apply #'urgrep query options)
        ;; Return the arguments to run directly.
-       (when (not (equal directory default-directory))
-         (error "Can't use plain urgrep with a non-default directory yet"))
        (unless (eshell-interactive-output-p)
-         (setq options (append '(:color nil) options)))
+         (setq options `(:color nil . ,options)))
        (throw 'eshell-replace-command
               (let* (;; Ensure we generate a POSIX shell-like command so that
                      ;; Eshell can (hopefully) parse it correctly.
